import Service from './service';
import User from './user';
import Message from './message';
import { CONST_SERVER_READY, CONST_SERVER_REGIST, CONST_SERVER_REGIST_SUCCESS, CONST_SERVER_REGIST_FAIL, CONST_SERVER_UNREGIST, } from './consts';
export class ServerMessage extends Message {
    constructor(from, to, title, body, reply, group) {
        super(from, to, title, body, reply, group);
    }
}
export default class MessageServer {
    constructor(userName) {
        this.usersMap = new Map();
        this.timeout = 10000;
        this._isStarted = false;
        this._userRegistedHandlesMap = new Map();
        this._registAdmin(userName);
        this.service = new Service();
        this.service.onMessage((event) => {
            var msg = event.data;
            if (!(msg && typeof msg.id == 'string' &&
                typeof msg.title == 'string' &&
                typeof msg.from == 'string' &&
                typeof msg.to == 'string')) {
                return;
            }
            switch (msg.title) {
                case CONST_SERVER_REGIST:
                    if (msg.to != this.agent.name) {
                        return;
                    }
                    var reason = this._validateRegist(msg.body);
                    if (reason) {
                        event.reply({ title: CONST_SERVER_REGIST_FAIL, body: reason });
                    }
                    else {
                        if (this.usersMap.has(userName) && this.usersMap.get(userName).address.closed) {
                            this.usersMap.get(userName).address = event.source;
                        }
                        else {
                            var user = new User(event.source, msg.body);
                            this.usersMap.set(user.name, user);
                        }
                        if (this._globalUserRegistedHandle) {
                            this._globalUserRegistedHandle.call(null, user.name);
                        }
                        if (this._userRegistedHandlesMap.get(user.name)) {
                            this._userRegistedHandlesMap.get(user.name).forEach(fn => fn(user.name));
                            this._userRegistedHandlesMap.delete(user.name);
                        }
                        event.reply({ title: CONST_SERVER_REGIST_SUCCESS, body: user.id });
                    }
                    break;
                case CONST_SERVER_UNREGIST:
                    if (msg.to != this.agent.name) {
                        return;
                    }
                    this.usersMap.delete(msg.body);
                    break;
            }
            if (msg.to != this.agent.name && this.usersMap.has(msg.from) && this.usersMap.has(msg.to)) {
                if (!msg.smid) {
                    let _msg = new ServerMessage(msg.from, msg.to, msg.title, msg.body, msg.reply, msg.group);
                    _msg.smid = _msg.id;
                    _msg.id = msg.id;
                    _msg.agent = this.agent.name;
                    this.postMessage(_msg);
                }
                return;
            }
        });
    }
    _validateRegist(userName) {
        if (typeof userName != 'string') {
            return 'userName must be string';
        }
        else if (userName.indexOf('@') == 0) {
            return 'userName can not begin with "@"';
        }
        else if (this.usersMap.has(userName) && !this.usersMap.get(userName).address.closed) {
            return 'username existed';
        }
        else {
            return;
        }
    }
    _registAdmin(userName = '@server') {
        var agent = new User(window, userName);
        agent.isAdmin = true;
        this.agent = agent;
    }
    start() {
        if (this._isStarted)
            return;
        this.service.start();
        this.broadcast({
            title: CONST_SERVER_READY,
        });
        return this;
    }
    stop() {
        this._isStarted = false;
        this.service.stop();
        return this;
    }
    restart() {
        this.service.restart();
        return this;
    }
    config(option) {
        this.service.config(option);
        return this;
    }
    onMessage(fn) {
        this.service.onMessage(fn);
        return this;
    }
    removeMessageHandler(fn) {
        this.service.removeMessageHandler(fn);
        return this;
    }
    postMessage(msg, needReplay = false) {
        return new Promise((resolve, reject) => {
            msg.from = msg.from || this.agent.name;
            if (this.usersMap.has(msg.to)) {
                var address = this.usersMap.get(msg.to).address;
                if (address && !address.closed) {
                    let msgId = this.service.postMessage(msg, undefined, undefined, address);
                    if (needReplay) {
                        let resolveTimer;
                        let replayHandle = (event) => {
                            let replyMsg = event.data;
                            if (replyMsg && replyMsg.reply == msgId) {
                                if (resolveTimer) {
                                    clearTimeout(resolveTimer);
                                    resolveTimer = undefined;
                                }
                                this.removeMessageHandler(replayHandle);
                                resolve(replyMsg);
                            }
                        };
                        this.onMessage(replayHandle);
                        resolveTimer = window.setTimeout(() => {
                            this.removeMessageHandler(replayHandle);
                            reject(new Error(`message from ${msg.from} to ${msg.to} title ${msg.title} timeout`));
                        }, this.timeout);
                    }
                }
                else {
                    reject(new Error(`user ${msg.to} address missing`));
                }
            }
            else {
                reject(new Error(`invalid address ${msg.to}`));
            }
        });
    }
    onUserRegisted(fn, userName) {
        if (!userName) {
            this._globalUserRegistedHandle = fn;
            return;
        }
        if (this.usersMap.has(userName)) {
            fn(userName);
        }
        else {
            if (!this._userRegistedHandlesMap.has(userName)) {
                this._userRegistedHandlesMap.set(userName, [fn]);
            }
            else {
                this._userRegistedHandlesMap.get(userName).push(fn);
            }
        }
    }
    broadcast(msg) {
        msg.from = this.agent.name;
        if (msg.to == '@all') {
            this.service.postMessage(Object.assign({}, msg));
        }
        else if (msg.to == '@allUser') {
            this.usersMap.forEach((user, key) => {
                this.service.postMessage(Object.assign(Object.assign({}, msg), { to: key }));
            });
        }
    }
}
