(function (factory) {
    if (typeof module === "object" && typeof module.exports === "object") {
        var v = factory(require, exports);
        if (v !== undefined) module.exports = v;
    }
    else if (typeof define === "function" && define.amd) {
        define(["require", "exports", "./service", "./consts"], factory);
    }
})(function (require, exports) {
    "use strict";
    Object.defineProperty(exports, "__esModule", { value: true });
    const service_1 = require("./service");
    const consts_1 = require("./consts");
    class MessageClient {
        constructor() {
            this.timeout = 10000;
            this._registHandle = this._registHandle.bind(this);
            this.service = new service_1.default();
            this.messageGroupMap = new Map();
        }
        registe(userName, serverName = '@server') {
            if (this._userName) {
                return Promise.reject('client registed');
            }
            if (this._registePromise) {
                return this._registePromise;
            }
            this._tempUserName = userName;
            this._serverName = serverName;
            if (!this._tempUserName || !this._serverName) {
                throw new Error('missing name and server');
            }
            this.service.start();
            this.service.onMessage(this._registHandle);
            this.service.postMessage({
                to: this._serverName,
                title: consts_1.CONST_SERVER_REGIST,
                body: this._tempUserName,
            });
            this._registePromise = new Promise((resolve, reject) => {
                this._registeResolve = resolve;
                this._registeReject = reject;
            });
            return this._registePromise;
        }
        unregiste() {
            if (this._userName) {
                this.postMessage({
                    title: consts_1.CONST_SERVER_UNREGIST,
                    body: this._userName,
                });
                this._userName = null;
                this._registePromise = null;
            }
            this.service.stop();
            return this;
        }
        start() {
            this.service.start();
            return this;
        }
        stop() {
            this.unregiste();
            return this;
        }
        restart() {
            this.service.restart();
            return this;
        }
        config(option) {
            this.service.config(option);
            return this;
        }
        onMessage(fn) {
            this.service.onMessage(fn);
            return this;
        }
        removeMessageHandler(fn) {
            this.service.removeMessageHandler(fn);
            return this;
        }
        postMessage(msg, needReplay = false) {
            if (!this._registePromise) {
                return this._postMessage(msg, needReplay = false);
            }
            return this._registePromise
                .then(() => {
                msg.from = this._userName;
                msg.to = msg.to || this._serverName;
                return this._postMessage(msg, needReplay = false);
            });
        }
        _postMessage(msg, needReplay = false) {
            return new Promise((resolve, reject) => {
                let msgId = this.service.postMessage(msg, undefined, undefined, this._serverAddress);
                if (needReplay) {
                    let resolveTimer;
                    let replayHandle = (event) => {
                        let replyMsg = event.data;
                        if (replyMsg && replyMsg.reply == msgId) {
                            if (resolveTimer) {
                                clearTimeout(resolveTimer);
                                resolveTimer = undefined;
                            }
                            this.removeMessageHandler(replayHandle);
                            resolve(replyMsg);
                        }
                    };
                    this.onMessage(replayHandle);
                    resolveTimer = window.setTimeout(() => {
                        this.removeMessageHandler(replayHandle);
                        reject(new Error(`message from ${msg.from} to ${msg.to} title ${msg.title} timeout`));
                    }, this.timeout);
                }
            });
        }
        _registHandle(event) {
            var msg = event.data;
            if (!(msg && typeof msg.id == 'string' &&
                typeof msg.title == 'string' &&
                typeof msg.from == 'string' &&
                typeof msg.to == 'string')) {
                return;
            }
            switch (msg.title) {
                case consts_1.CONST_SERVER_REGIST_SUCCESS:
                    if (this._registeResolve && msg.from == this._serverName) {
                        this._userName = this._tempUserName;
                        this._serverAddress = event.source;
                        this._registeResolve(msg.body);
                        this.service.removeMessageHandler(this._registHandle);
                    }
                    break;
                case consts_1.CONST_SERVER_REGIST_FAIL:
                    if (this._registeReject && msg.from == this._serverName) {
                        this._registeReject(msg.body);
                        this.service.removeMessageHandler(this._registHandle);
                    }
                case consts_1.CONST_SERVER_READY:
                    if (this._registePromise && !this._userName) {
                        this.service.postMessage({
                            to: this._serverName,
                            title: consts_1.CONST_SERVER_REGIST,
                            body: this._tempUserName,
                        });
                    }
                    break;
            }
        }
    }
    exports.default = MessageClient;
});
