import Service, { IMessageHandle, IMsgObj, IServiceOption, CrossAppCommunicationEvent } from './service';
import {
    CONST_SERVER_READY,
    CONST_SERVER_REGIST,
    CONST_SERVER_REGIST_SUCCESS,
    CONST_SERVER_REGIST_FAIL,
    CONST_SERVER_UNREGIST,
} from './consts';

export default class MessageClient {
    service: Service;

    messageGroupMap: Map<string, IMessageHandle>;

    timeout: number = 10000;

    private _serverName: string;

    private _serverAddress: Window;

    private _tempUserName: string;

    private _userName: string;

    private _registePromise: Promise<any>;

    private _registeResolve: any;

    private _registeReject: any;

    constructor() {
        this._registHandle = this._registHandle.bind(this);
        this.service = new Service();
        this.messageGroupMap = new Map();
    }

    registe(userName: string, serverName = '@server') {
        if (this._userName) {
            return Promise.reject('client registed');
        }

        if (this._registePromise) {
            return this._registePromise;
        }

        this._tempUserName = userName;
        this._serverName = serverName;

        if (!this._tempUserName || !this._serverName) {
            throw new Error('missing name and server');
        }

        this.service.start();
        this.service.onMessage(this._registHandle);
        this.service.postMessage({
            to: this._serverName,
            title: CONST_SERVER_REGIST,
            body: this._tempUserName,
        });

        this._registePromise = new Promise((resolve, reject) => {
            this._registeResolve = resolve;
            this._registeReject = reject;
        });

        return this._registePromise;
    }

    unregiste() {
        if (this._userName) {
            this.postMessage({
                title: CONST_SERVER_UNREGIST,
                body: this._userName,
            });
            this._userName = null;
            this._registePromise = null;
        }
        this.service.stop();
        return this;
    }

    start() {
        this.service.start();
        return this;
    }

    stop() {
        this.unregiste();
        return this;
    }

    restart() {
        this.service.restart();
        return this;
    }

    config(option: IServiceOption) {
        this.service.config(option);
        return this;
    }

    onMessage(fn: IMessageHandle) {
        this.service.onMessage(fn);
        return this;
    }

    removeMessageHandler(fn: IMessageHandle) {
        this.service.removeMessageHandler(fn);
        return this;
    }

    postMessage(msg: IMsgObj, needReplay = false) {
        if (!this._registePromise) {
            return this._postMessage(msg, needReplay = false);
        }

        return this._registePromise
            .then(() => {
                msg.from = this._userName;
                msg.to = msg.to || this._serverName;
                return this._postMessage(msg, needReplay = false);
            });
    }

    _postMessage(msg: IMsgObj, needReplay = false) {
        return new Promise((resolve, reject) => {
            let msgId = this.service.postMessage(msg, undefined, undefined, this._serverAddress);
            if (needReplay) {
                let resolveTimer: number;
                let replayHandle = (event: CrossAppCommunicationEvent) => {
                    let replyMsg = event.data;
                    if (replyMsg && replyMsg.reply == msgId) {
                        if (resolveTimer) {
                            clearTimeout(resolveTimer);
                            resolveTimer = undefined;
                        }
                        this.removeMessageHandler(replayHandle);
                        resolve(replyMsg);
                    }
                }
                this.onMessage(replayHandle);
                resolveTimer = window.setTimeout(() => {
                    this.removeMessageHandler(replayHandle);
                    reject(new Error(`message from ${msg.from} to ${msg.to} title ${msg.title} timeout`))
                }, this.timeout);
            }
        })
    }

    _registHandle(event: CrossAppCommunicationEvent) {
        var msg = event.data;

        if (!(msg && typeof msg.id == 'string' &&
            typeof msg.title == 'string' &&
            typeof msg.from == 'string' &&
            typeof msg.to == 'string')) {
            return;
        }

        switch (msg.title) {
            case CONST_SERVER_REGIST_SUCCESS:
                if (this._registeResolve && msg.from == this._serverName) {
                    this._userName = this._tempUserName;
                    this._serverAddress = event.source;
                    this._registeResolve(msg.body);
                    this.service.removeMessageHandler(this._registHandle);
                }
                break;
            case CONST_SERVER_REGIST_FAIL:
                if (this._registeReject && msg.from == this._serverName) {
                    this._registeReject(msg.body);
                    this.service.removeMessageHandler(this._registHandle);
                }
            case CONST_SERVER_READY:
                if (this._registePromise && !this._userName) {
                    this.service.postMessage({
                        to: this._serverName,
                        title: CONST_SERVER_REGIST,
                        body: this._tempUserName,
                    });
                }
                break;
        }
    }
}