/**
 * 基于云桌面应用间通信协议实现的基础通信服务
 */

import Message from './message';

export interface IMsgObj {
    /** 回复消息id，如果是回复某条消息，则有此属性 */
    reply?: string,

    /**
     * 消息发送方的标识符，默认为location.href
     */
    from?: string,

    /** 消息接收方的标识符，默认为'_all_' */
    to?: string,

    /** 消息主题 */
    title: string,

    /** 消息内容 */
    body?: any,

    /** 消息组 */
    group?: string,
}

export interface IMessageHandle {
    (event: CrossAppCommunicationEvent): any;
}

/**
 * 向所有有关系的窗口发送消息
 * @params {Message} msg 消息对象
 * @params {string} targetOrigin
 * @params {Transferable[]} transfer
 * @params {Window} [address] 接收消息的窗口，如果不传，则进行广播，发到所有能发的窗口
 */
function mssageSender(msg: Message, targetOrigin: string, transfer: any[], address?: Window) {
    try {
        if (address && typeof address.postMessage == 'function') {
            address.postMessage(msg, targetOrigin, transfer);
            return;
        }

        var ifrs = document.getElementsByTagName('iframe');
        Array.prototype.forEach.call(ifrs, (ifr: HTMLIFrameElement) => {
            ifr.contentWindow.postMessage(msg, targetOrigin, transfer);
        });
        if (window.parent != window) {
            window.parent.postMessage(msg, targetOrigin, transfer);
        }
        if (window.opener != null) {
            window.opener.postMessage(msg, targetOrigin, transfer);
        }
        window.postMessage(msg, targetOrigin, transfer);
    } catch (error) {
        console.error(error);
    }
}

export interface IFilter {
    (msg: Message): boolean;
}

function defaultFilter(msg: any) { if (msg && msg.id && msg.from && msg.to && msg.title) return true }

export interface IServiceOption {
    filter?: IFilter;
    maxHistory?: number;
}

export class CrossAppCommunicationEvent {
    public orginEvent: MessageEvent;
    public data: Message;
    public origin: string;
    public source: any;
    public msgId: string;
    public from: string;
    public to: string;
    public group: string | undefined;

    constructor(event: MessageEvent) {
        this.orginEvent = event;
        this.data = event.data;
        this.origin = event.origin;
        this.source = event.source;

        this.msgId = event.data.id;
        this.from = event.data.from;
        this.to = event.data.to;
        this.group = event.data.group;
    }

    /**
     * 快捷回复消息接口
     * @params {MsgObj} msg 消息对象
     * @params {Transferable[]} 转交给消息接收方的特殊对象(ArrayBuffer, MessagePort and ImageBitmap)，发送方将不可时用这些对象了。
     */
    reply(msg: IMsgObj, transfer?: any[]) {
        try {
            msg = new Message(msg.from || this.to || location.href, this.from, msg.title, msg.body, this.msgId, this.group);
        } catch (error) {
            console.error(error);
        }
        if (!(msg instanceof Message)) {
            return;
        }

        this.source.postMessage(msg, this.origin, transfer);
    }

    destroy() {
        this.orginEvent = undefined;
        this.data = undefined;
        this.origin = undefined;
        this.source = undefined;

        this.msgId = undefined;
        this.from = undefined;
        this.to = undefined;
    }
}

export default class CrossAppCommunicationService {
    private _isStarted: boolean = false;
    private _listeners: IMessageHandle[] = [];
    private _history: Message[] = [];
    public option: IServiceOption = {
        filter: defaultFilter,
        maxHistory: 100,
    };

    constructor() {
        this._onMessageHandle = this._onMessageHandle.bind(this);
    }

    start(): CrossAppCommunicationService {
        if (this._isStarted) {
            return;
        }

        window.addEventListener("message", this._onMessageHandle);
        this._isStarted = true;

        return this;
    }

    stop(): CrossAppCommunicationService {
        window.removeEventListener("message", this._onMessageHandle);
        this._listeners = [];
        this._history = [];
        this._isStarted = false;

        return this;
    }

    restart(): CrossAppCommunicationService {
        return this.stop().start();
    }

    config(option: IServiceOption): CrossAppCommunicationService {
        this.option = Object.assign(this.option, option);

        return this;
    }

    /**
     * 发送消息接口
     * @params {MsgObj|Message} option 消息对象
     * @params {string} option.from 消息发送方的应用实例Id, iframe型应用和desktop可省略
     * @params {string} option.to 消息接收方的应用实例Id, iframe型应用可省略
     * @params {string} option.title 消息主题
     * @params {any} option.body 发送的消息内容, 空消息可发送'',null,undefined
     * @params {number} [option.reply] 回复消息id，如果是回复某条消息，则有此属性
     * @params {string} [targetOrigin] 消息接收方的域，默认为'*', 如果发送的消息中包含密码等信息，最好制定此参数
     * @params {Transferable[]} [transfer] 转交给消息接收方的特殊对象(ArrayBuffer, MessagePort and ImageBitmap)，发送方将不可时用这些对象了。
     * @params {Window} [address] 指定的消息发送地址，window对象
     */
    postMessage(msg: IMsgObj | Message, targetOrigin?: string, transfer?: any[], address?: Window): string {
        var _msg: Message;
        if (!(msg instanceof Message)) {
            try {
                msg.from = msg.from || location.href;
                msg.to = msg.to || '@all';
                _msg = new Message(msg.from, msg.to, msg.title, msg.body, msg.reply, msg.group);
            } catch (error) {
                console.error(error);
            }
            if (!(_msg instanceof Message)) {
                return;
            }
        } else {
            _msg = msg;
        }

        targetOrigin = targetOrigin || '*';
        transfer = transfer || [];

        mssageSender(_msg, targetOrigin, transfer, address);

        this._logHistory(_msg);

        return _msg.id;
    }

    /**
     * 监听消息消息接口
     * @params {MessageHandle} fn 消息发送方的应用实例Id, iframe型应用和desktop可省略
     */
    onMessage(fn: IMessageHandle): CrossAppCommunicationService {
        if (typeof fn != 'function') {
            console.error('onMessage arguments must be a function');
            return;
        }

        var _listeners = this._listeners;

        if (_listeners.indexOf(fn) == -1) {
            _listeners.push(fn);
        }

        return this;
    }

    removeMessageHandler(fn: IMessageHandle) {
        if (typeof fn != 'function') {
            console.error('onMessage arguments must be a function');
            return;
        }

        var _listeners = this._listeners;
        var index = _listeners.indexOf(fn);

        if (index != -1) {
            _listeners.splice(index, 1);
        }
    }

    private _logHistory(msg: Message) {
        if (this._history.length < this.option.maxHistory) {
            this._history.push(msg);
        } else {
            this._history.splice(0, 1, msg);
        }
    }

    private _onMessageHandle(event: MessageEvent) {
        var msg = event.data;
        var filter = this.option.filter || defaultFilter;

        if (filter(msg)) {
            var customEvent = new CrossAppCommunicationEvent(event);
            var _listeners = this._listeners;

            this._logHistory(msg);

            _listeners.forEach(listener => listener(customEvent));
        }
    }
}